//
//	$Date: 2001/03/08 03:16:55 $
//	$Revision: 1.3 $
//
//	Aho-Corasick$BK!$K$h$k%Q%?%s%^%C%A(B
//
//	$B:G=i$+$i%H%i%$$,M-$k$J$i:n$j$J$*$9I,MW$b$J$$$+$b(B
//
#include <Pilot.h>

#ifndef NULL
#define NULL 0
#endif

#include "acsearch.h"

#define UNDEF 0xffff
#define FAIL 0xfffe

#define MAXNODES 60

//static unsigned short trans[MAXNODES][0x100];
//static unsigned short _trans[MAXNODES * 0x100];
static unsigned short *_trans = NULL;
#define trans(x,y) _trans[((x)<<8)+y]
static unsigned char terminal[MAXNODES];
static unsigned char character[MAXNODES];
static unsigned short parent[MAXNODES];
static unsigned short failure[MAXNODES];
static unsigned short nnodes;

static void
init()
{
	register int i,j;
	if(_trans == NULL){
		_trans = (unsigned short*)MemPtrNew(sizeof(short)*MAXNODES*0x100);
		ErrFatalDisplayIf(_trans==NULL,"Trans");
	}
	for(i=0;i<MAXNODES;i++){
		for(j=0;j<0x100;j++){
			trans(i,j) = FAIL;
		}
		terminal[i] = 0;
		character[i] = 0;
		failure[i] = UNDEF;
	}
	nnodes = 1;
}

static int
regpat(int n, unsigned char *s)
{
	unsigned short node;
	if(nnodes >= MAXNODES) return -1;
	if(*s == NULL || *s == '\r' || *s == '\n'){
		terminal[n] = 1;
		return 0;
	}
	node = trans(n,*s);
	if(node == FAIL){
		node = nnodes++;
		if(node >= MAXNODES) return -1;
		character[node] = *s;
		parent[node] = n;
		trans(n,*s) = node;
	}
	return regpat(node,s+1);
}

static int
calcf(int n)
{
	register int c,r;
	if(failure[n] != UNDEF) return failure[n];
	c = character[n];
	r = calcf(parent[n]);
	while(r != FAIL && trans(r,c) == FAIL){
		r = calcf(r);
	}
	if(r == FAIL){
		failure[n] = trans(0,c);
	}
	else {
		failure[n] = trans(r,c);
		if(terminal[failure[n]]){
			terminal[n] = 1;
		}
	}	
}

static void
calcfailure()
{
	register int i;
	unsigned short n;

	failure[0] = FAIL;
	for(i=0;i<0x100;i++){
		n = trans(0,i);
		if(n != FAIL){
			failure[n] = 0;
		}
	}
	for(i=1;i<nnodes;i++){
		calcf(i);		
	}
}

static void
calctrans()
{
	register int i,j,n;

	for(i=0;i<nnodes;i++){
		for(j=0;j<0x100;j++){
//			if(j < 0x20 && j != '\r' && j != '\n' && j != '\t') continue;
			if(trans(i,j) != FAIL) continue;
			n = failure[i];
			while(n != FAIL && trans(n,j) == FAIL){
				n = failure[n];
			}
			trans(i,j) = (n != FAIL ? trans(n,j) :
				       trans(0,j) == FAIL ? 0 :
				       trans(0,j));
		}
	}
/*
	for(i=0;i<nnodes;i++){
		for(j=0;j<0x100;j++){
			if(trans[i][j] < 0  || trans[i][j] >= nnodes){
				DumpHex(i,10,33);
				DumpHex(j,10,63);
				DumpHex(trans[i][j],10,93);
			}
		}
	}
*/
}

void
acsearch_beginpat()
{
	init();
}

int
acsearch_addpat(register unsigned char *s)
{
	register unsigned char *p = s;
	while(*p){
		while(*p == '\r' || *p == '\n') p++;
		if(regpat(0,p) < 0) return -1;
		while(*p != '\r' && *p != '\n' && *p) p++;
	}
	return 0;
}

void
acsearch_endpat()
{
	calcfailure();
	calctrans();
}

int
acsearch_makepat(register unsigned char *s)
{
	register unsigned char *p = s;
	init();
	while(*p){
		while(*p == '\r' || *p == '\n') p++;
		if(regpat(0,p) < 0) return -1;
		while(*p != '\r' && *p != '\n' && *p) p++;
	}
	calcfailure();
	calctrans();
	return 0;
}

int
acsearch_match(register unsigned char *text)
{
	register unsigned short state = 0;
	for(;*text;text++){
		state = trans(state,*text);
		if(terminal[state]) return 1;
	}
	return 0;
}

int
acsearch_match_size(register unsigned char *text, register int size)
{
	unsigned short state = 0;
	for(;size > 0;text++,size--){
		state = trans(state,*text);
		if(terminal[state]) return 1;
	}
	return 0;
}

#ifdef TEST
#include <stdio.h>

main()
{
	int i,j;

//	acsearch_makepat("aba\nbac\nbbc\ncab");

	acsearch_beginpat();
//	acsearch_addpat("aba");
//	acsearch_addpat("bac");
//	acsearch_addpat("bbc");
//	acsearch_addpat("cab");
	acsearch_addpat("ha");
	acsearch_endpat();

	{
		char buf[100];
		while(gets(buf)){
			if(acsearch_match(buf)){
				printf("match!\n");
			}
		}
	}
}
#endif

#ifdef DEBUG
traverse1(int n, int level)
{
	int i,j;
	if(terminal[n]){
		for(i=0;i<level;i++) printf("\t");
	}
	for(j=0;j<0x100;j++){
		if(trans[n][j] != FAIL){
			for(i=0;i<level;i++) printf("\t");
			printf("[%02x]\n",j);
			traverse1(trans[n][j],level+1);
		}
	}
}

traverse(int n)
{
	traverse1(n,0);
}
#endif

